/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <fcntl.h>
#include "is_types.h"
#include "isam.h"

#ifndef lint
static	char	sccsid[] = "@(#)db_write.c	1.10	11:48:41	2/7/92";
#endif

int	isam_add();
unsigned long	isam_serial();

extern	int	db_errno;

/*
 * __db_write_key - add a new key for the new record
 *
 * __db_write_key() assembles an index key from the record which is to be
 * added to the database.  That index key is added to the selected index.
 */

int
__db_write_key (file, index, data, record)
file_t	*file;
int	index;
void	*data;
off_t	record;
{
	char	buf[MAXBLOCKSIZE];
	sort_key_t *sort_key = &file->f_index->i_keys[index];

	(void) __isam_buf2key (file, index, data, record, buf);
	return isam_add (file, index, buf, record);
}

/*
 * __db_write_record - add a new record to a database file
 *
 * __db_write_record() adds the data for a new record and adds all of
 * the index keys for the indices defined for the file.
 */

int
__db_write_record (file, data)
file_t	*file;
char	*data;
{
	int	i;
	off_t	record = -1;
	field_t	*fields = file->f_schema->s_fields;
	int	nfields = file->f_schema->s_field_count;
	int	was_locked = file->f_locked;
	long	ticket;

	/*
	 * See if there is a serial number field in this file.  If there
	 * is, we will have to get a serial number for this record.  We
	 * do this first to avoid the hassle of locking and unlocking the
	 * file in case of failure.
	 */

	for (file->f_serial = 0L, i = 0;i < nfields;i++) {
		if (fields[i].f_type == T_SERIAL) {
			file->f_serial = isam_serial (file);
			if (file->f_serial == (unsigned long) -1L)
				return -1;

			memcpy (data + fields[i].f_offset,
				&file->f_serial, sizeof file->f_serial);

			break;
		}
	}

	/*
	 * Lock the file to keep the indices from changing out from
	 * under your feet.
	 */

	if (was_locked != F_WRLCK)
		if ((ticket = isam_set_ticket (file)) == -1)
			return -1;

	/*
	 * Allocate a new data record and write it out.  The record number
	 * is required for non-unique indicies so we must do this first.
	 */

	if ((record = __db_alloc (file)) == -1)
		goto fail;

	if (isam_write_data (file, record, data))
		goto fail;

	/*
	 * Now write out all the keys.  In order to preserve the current
	 * key, all of the other keys are written out, then the current
	 * key is written.  This preserves all of the key pointers so that
	 * the isam_next and isam_previous functions will work after a
	 * record is added.
	 */

	for (i = 1;i <= file->f_index->i_key_count;i++) {
		if (i == file->f_current_index)
			continue;

		if (__db_write_key (file, i, data, record))
			goto un_index;
	}
	if (file->f_current_index != 0 && __db_write_key (file,
			file->f_current_index, data, record))
		goto un_index;

	/*
	 * The indices are finished being modified, so unlock the
	 * file.  Mark the new record as current, and return success.
	 */

	if (! was_locked) {
		file->f_ticket = ticket;
		isam_end_ticket (file);
	}
	file->f_current_record = record;
	return 0;

fail:
	/*
	 * Something failed, so the freshly allocated data block must be
	 * freed.  Any lock which was held needs to be released.
	 */

	if (record != -1)
		__db_free (file, record);

	if (! was_locked) {
		file->f_ticket = ticket;
		isam_end_ticket (file);
	}
	return -1;

un_index:

	/*
	 * Undo the indices that were updated.  The failing index
	 * would have been the last one, so we decrement the index
	 * number.  If all of the non-current indices were updated,
	 * `i' would be one too big, so this works both ways.  Also,
	 * it is impossible to fail updating the current index and
	 * still have to remove it, so never delete the key for the
	 * current index.
	 */

	for (i--;i >= 1;i--) {
		if (i != file->f_current_index)
			(void) __db_delete_key (file, i, data, record);
	}
	goto fail;
}

/*
 * db_write_record - add a new database record
 *
 * db_write_record() validates the input parameters and invokes the
 * helper routine __db_write_record, which performs the actual work.
 */

int
db_write_record (dbm, filenum, buf)
dbm_t	*dbm;
int	filenum;
VOIDP	*buf;
{
	file_t	*file;

	/*
	 * Insure that the file number is in the proper range
	 */

	if (filenum < 0 || filenum >= dbm->d_file_count ||
			dbm->d_files[filenum].f_file_name[0] == '\0')
		return (db_errno = DB_INVALID), -1;

	/*
	 * Get the pointer to the file and check that it has
	 * actually been opened.  Then invoke __db_write_record
	 * to do the actual work.
	 */

	file = &dbm->d_files[filenum];
	if (file->f_schema->s_fields == 0)
		return (db_errno = DB_NOTOPEN), -1;

	return __db_write_record (file, (char *) buf);
}
