/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#ifdef	BSD
#include <strings.h>
#define	strchr	index
#define	strrchr	rindex
#else
#include <string.h>
#endif
#ifdef MALLOC_H
#include <malloc.h>
#endif
#include "is_types.h"
#include "is_proto.h"
#include "isam.h"
#include "patchlevel.h"

#ifndef lint
static	char	sccsid[] = "@(#)dbload.c	1.9	09:21:23	3/30/92";
#endif

dbm_t	*db_create();
dbm_t	*db_open();
int	db_open_file();
int	db_create_file();
int	db_create_field();

#ifndef	MALLOC_H
#ifdef	__STDC__
void	*malloc();
#else
char	*malloc();
#endif
#endif

extern	int	db_errno;

/*
 * usage - display a usage message and exit
 */

usage ()
{
	fprintf (stderr, "usage: dbload [ -b ][ -d delimiter ] [ -f input ] database file\n");
	exit (1);
}

/*
 * dbload - read a text file and convert it to data records
 *
 * dbload converts text records into binary data records which are added
 * to the named file.  The syntax for the dbload command is
 *
 *	dbload [ -b ] [ -d c ] [ -f input ] database file
 *
 * The -b option is used to select "binary" input mode.  The input stream
 * is treated like binary records which are already in the correct format.
 *
 * The -d option is used to select the delimiter character.  By default
 * the delimiter character is "|".
 *
 * The -f option is used to select a file to be read as input.  By default
 * standard input is used.
 */

main (argc, argv)
int	argc;
char	**argv;
{
	dbm_t	*dbm;
	file_t	*file;
	field_t	*fields;
	int	filenum;
	char	*buf;
	char	*record;
	char	*cp;
	char	error_msg[BUFSIZ];
	field_t	new_field;
	int	field;
	int	i, j;
	int	delimiter = '|';
	int	binary = 0;
	int	flag;
	int	size;
	long	recno = 0;
	extern	int	optind;
	extern	char	*optarg;

	/*
	 * Loop over the command line arguments
	 */

	while ((flag = getopt (argc, argv, "bd:f:V")) != EOF) {
		if (flag == 'd') {
			delimiter = *optarg;
		} else if (flag == 'b') {
			binary = 1;
		} else if (flag == 'f') {

			/*
			 * Replace stdin with the named file
			 */

			if (freopen (optarg, "r", stdin) == 0) {
				fprintf (stderr, "dbload: can't open %s",
					optarg);
				exit (1);
			}
		} else if (flag == 'V') {
			fprintf (stderr, "dbload: version %s\n", VERSION);
			exit (1);
		} else
			usage ();
	}

	/*
	 * There must always exactly two arguments left - the name of
	 * the database to use and the name of the file in that database.
	 */

	if (optind + 2 != argc)
		usage ();

	/*
	 * Open the requested database and file.
	 */

	if (! (dbm = db_open (argv[optind], O_RDWR))) {
		sprintf (error_msg, "dbload: can't open database %s",
			argv[optind]);
		db_error (error_msg);
		exit (1);
	}
	if ((filenum = db_open_file (dbm, argv[optind + 1])) < 0) {
		sprintf (error_msg, "dbload: can't open file %s",
			argv[optind + 1]);
		db_error (error_msg);
		db_close (dbm);
		exit (1);
	}

	/*
	 * Allocate space to hold a single data record.  The input record
	 * will be placed here.  The input record must be large enough to
	 * hold the largest possible record.  For binary mode, it is just
	 * the record size, but in text mode it is up to 5 times the size.
	 */

	size = dbm->d_files[filenum].f_record_size;
	record = malloc (size);

	if (binary)
		buf = malloc (size);
	else
		buf = malloc (5 * size);

	if (buf == 0) {
		perror ("dbload: no memory");
		db_close (dbm);
		exit (1);
	}

	/*
	 * Lock the file to save on ticket taking and improve performance
	 * with ISAM block buffering.
	 */

	if (db_lock (dbm, filenum, 1)) {
		sprintf (error_msg, "dbload: can't lock %s",
			argv[optind + 1]);
		db_error (error_msg);
		db_close (dbm);
		exit (1);
	}
	if (isam_setbuf (100)) {
		db_error ("dbload: can't set file buffering");
		db_unlock (dbm, filenum);
		db_close (dbm);
		exit (1);
	}

	/*
	 * In binary mode we read a record and add it to the database
	 * as is.
	 */

	if (binary) {
		while (read (fileno (stdin), record, size) == size) {
			recno++;
			if (db_write_record (dbm, filenum, record)) {
				sprintf (error_msg, "dbload: error writing record %ld",
					recno);
				db_error (error_msg);
			}
		}
	} else {

	/*
	 * In text mode we read a line and convert the line to binary
	 * format.  That buffer is then added to the database file.
	 */

		while (fgets (buf, size * 5, stdin)) {

			if (cp = strchr (buf, '\n'))
				*cp = '\0';

			recno++;

			if (string2buf (&dbm->d_files[filenum],
					buf, delimiter, record)) {
				if (db_write_record (dbm, filenum, record)) {
					sprintf (error_msg,
						"dbload: error writing record %ld",
						recno);
					db_error (error_msg);
				}
			} else {
				fprintf (stderr, "dbload: error converting record %ld\n",
					recno);
			}
		}
	}
	db_unlock (dbm, filenum);
	db_close (dbm);
	exit (0);
}
